"""
* Filename: core.py
* Description: cli core function, processing various requests
* Time: 2020.11.30
* Author: liuf5
*/
"""

import re
import os
from zoomeye import config, file, show
from zoomeye.sdk import ZoomEye
from zoomeye.data import CliZoomEye, HistoryDevice, IPInformation, DomainSearch

# save zoomeye config folder
zoomeye_dir = os.path.expanduser(config.ZOOMEYE_CONFIG_PATH)


def key_init(key):
    """
    initialize through the api key, write the api key to the local configuration file,
    theoretically it will never expire unless you remake the api key
    :param key: user input API key
    :return:
    """
    file.check_exist(zoomeye_dir)
    key = key.strip()
    try:
        zoom = ZoomEye(api_key=key)
    except Exception:
        return
    # api key save path
    key_file = zoomeye_dir + "/apikey"
    # display the remaining resources of the current account
    user_data = zoom.resources_info()
    show.printf("Role: {}".format(user_data["plan"]))
    show.printf("Quota: {}".format(user_data["resources"].get("search")))
    # save api key
    with open(key_file, 'w') as f:
        f.write(key)
    show.printf("successfully initialized", color="green")
    # change the permission of the configuration file to read-only
    os.chmod(key_file, 0o600)


def jwt_init(username, password):
    """
    initialize through the user name and password, write jwt to the local configuration file,
    the expiration time is about 12 hours, so it is recommended to initialize through the api key.
    :param username: str, login zoomeye account
    :param password: str, login zoomeye account password
    :return:
    """
    file.check_exist(zoomeye_dir)
    try:
        zoom = ZoomEye(username=username, password=password)
        access_token = zoom.login()
    except Exception:
        return
    jwt_file = zoomeye_dir + "/jwt"
    if access_token:
        # display the remaining resources of the current account
        user_data = zoom.resources_info()
        show.printf("Role: {}".format(user_data["plan"]))
        show.printf("Quota: {}".format(user_data["resources"].get("search")))
        with open(jwt_file, 'w') as f:
            f.write(access_token)
        show.printf("successfully initialized", color="green")
        # change the permission of the configuration file to read-only
        os.chmod(jwt_file, 0o600)
    else:
        show.printf("failed initialized!", color="red")


def init(args):
    """
    the initialization processing function will select the initialization method according to the user's input.
    :param args:
    :return:
    """
    api_key = args.apikey
    username = args.username
    password = args.password
    # use api key init
    if api_key and username is None and password is None:
        key_init(api_key)
        return
    # use username and password init
    if api_key is None and username and password:
        jwt_init(username, password)
        return
    # invalid parameter
    show.printf("input parameter error", color="red")
    show.printf("please run <zoomeye init -h> for help.", color="red")


def search(args):
    dork = args.dork
    num = int(args.num)
    facet = args.facet
    filters = args.filter
    stat = args.stat
    save = args.save
    count_total = args.count
    figure = args.figure
    force = args.force
    resource = args.type

    cli_zoom = CliZoomEye(dork, num, resource=resource, facet=facet, force=force)
    if filters:
        cli_zoom.filter_data(filters, save)
        return
    if facet:
        cli_zoom.facets_data(facet, figure)
        return
    if count_total:
        cli_zoom.count()
        return
    if stat:
        cli_zoom.statistics(stat, figure)
        return
    if save:
        cli_zoom.save(save)
        return
    if filters is None and facet is None and stat is None:
        cli_zoom.default_show()
        return
    show.printf("please run <zoomeye search -h> for help.")


def info(args):
    """
    used to print the current identity of the user and the remaining data quota for the month
    :param args:
    :return:
    """
    api_key, access_token = file.get_auth_key()
    zm = ZoomEye(api_key=api_key, access_token=access_token)
    # get user information
    user_data = zm.resources_info()
    if user_data:
        # show in the terminal
        show.printf("Role: {}".format(user_data["plan"]))
        show.printf("Quota: {}".format(user_data["resources"].get("search")))


def ip_history(args):
    """
    query device history
    please see: https://www.zoomeye.org/doc#history-ip-search
    :param args:
    :return:
    """
    ip = args.ip
    filters = args.filter
    force = args.force
    number = args.num
    # determine whether the input is an IP address by regular
    compile_ip = re.compile('^(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|[1-9])\.(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|\d)\.'
                             '(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|\d)\.(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|\d)$')
    # IP format error，exit program
    if not compile_ip.match(ip):
        show.printf("[{}] it is not an IP address, please check!".format(ip), color='red')
        return

    zm = HistoryDevice(ip, force, number)
    # user input filter field
    if filters:
        filter_list = filters.split(',')
        zm.filter_fields(filter_list)
        return
    # no filter field,
    # print [timestamp,service,port,country,raw_data] fields
    zm.show_fields()


def clear_file(args):
    """
    clear user setting and zoomeye cache data
    """
    setting = args.setting
    cache = args.cache
    target_dir = None
    # clear user setting
    if setting:
        target_dir = zoomeye_dir
    # clear local cache file
    if cache:
        target_dir = os.path.expanduser(config.ZOOMEYE_CACHE_PATH)
    # user input error
    if target_dir is None:
        show.printf("Please run <zoomeye clear -h> for help!", color='red')
        return
    # remove all files under the folder
    file_list = os.listdir(target_dir)
    for item in file_list:
        os.remove(os.path.join(target_dir, item))
    show.printf("clear complete!", color='green')


def information_ip(args):
    ip = args.ip
    filters = args.filter
    # determine whether the input is an IP address by regular
    compile_ip = re.compile('^(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|[1-9])\.(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|\d)\.'
                            '(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|\d)\.(1\d{2}|2[0-4]\d|25[0-5]|[1-9]\d|\d)$')
    # IP format error，exit program
    if not compile_ip.match(ip):
        show.printf("[{}] it is not an IP address, please check!".format(ip), color='red')
        return

    infor = IPInformation(ip)
    if filters:
        filter_list = filters.split(',')
        infor.filter_information(filter_list)
        return

    infor.show_information()


def associated_domain_query(args):
    q = args.q
    resource = args.type
    page = args.page
    # show information for user
    DomainSearch(q, resource, page).show_information()
    return None




"""
* Filename: show.py
* Description: used to print different data on the terminal.
* Time: 2020.11.30
* Author: liuf5
*/
"""
import re
import operator
from colorama import init
from zoomeye import config, data, plotlib
from zoomeye.sdk import ZoomEyeDict

# solve the color display error under windows terminal
init(autoreset=True)


def convert_str(s):
    """
    convert arbitrary string to another string, for print and human readable
    :param:
    :return:
    """
    res = []
    d = {
        '\n': '\\n',
        '\r': '\\r',
        '\t': '\\t',
        '\b': '\\b',
        '\a': '\\a',
    }
    for c in s:
        if c in d.keys():
            res.append(d[c])
        else:
            res.append(c)
    return ''.join(res)


def omit_str(text, index=0):
    """
    ignore part of the string
    :param text: str, text to be omitted
    :param index: int, position omitted on the longest basis
    :return: str, ex: string...
    """
    if len(text) < config.STRING_MAX_LENGTH:
        return text

    return text[:config.STRING_MAX_LENGTH - index] + '...'


def printf(s, color="white"):
    """
    print result in terminal
    :param s: str, printed string
    :param color: str, printed string color
    :return:
    """
    colors = {
        "red":      '\033[31m',  # red
        "green":    '\033[32m',  # green
        "yellow":   '\033[33m',  # yellow
        "blue":     '\033[34m',  # blue
        "white":    '\033[37m',  # white
    }

    # default color
    default = '\033[0m'  # item default color
    print('{}{}{}'.format(colors.get(color, "WHITE"), s, default))


def show_host_default_data(data_list, count):
    """
    display the dork data searched by the user on the terminal,
    if facet is specified, it will also be displayed.
    :param resource: search type web or host, str
    :param data_list: dork data , list
    :param count: API data count , int
    :return:
    """
    total = 0
    # print title
    printf("{:<25}{:<20}{:<20}{:<30}{:<30}".format(
        "ip:port", "service", "country", "app", "banner"), color="green")

    # process data
    for d in data_list:
        ip, country, port, app, banner, service = data.get_host_item(d)
        # content length too long, truncated
        app_name = omit_str(app, 5)

        # str --> hex
        banner_content = omit_str(convert_str(omit_str(banner)))

        # print result
        total += 1
        printf('{:<25}{:<20}{:<20}{:<30}{:30}'.format(
            "{}:{}".format(ip, port), service, country, app_name,
            banner_content)
        )
    print()
    printf("total: {}/{}".format(total, count))


def show_web_default_data(data_list, count):
    # print title
    printf("{:<27}{:<27}{:<27}{:<30}".format(
         "site", "title", "country", "banner"), color="green")
    num = 0
    for item in data_list:
        num += 1
        item_dict = ZoomEyeDict(item)
        content = ''
        for item_key in data.default_table_web.keys():
            dict_result = item_dict.find(data.default_table_web.get(item_key))
            if isinstance(dict_result, list):
                for result_item in dict_result:
                    if isinstance(result_item, dict):
                        text = result_item.get("name")
                    elif isinstance(result_item, str):
                        text = result_item
                    else:
                        text = 'unknown'
            else:
                text = dict_result
            if text is None:
                text = 'unknown'
            text = convert_str(omit_str(text))
            content += "{:<27}".format(text)
        printf(content)
    printf('')
    printf("total:{}/{}".format(num, count))


def print_filter(keys, data_list, condition=None):
    """
    used to display user filtered data on the terminal
    :param keys: user input key, is str
    :param data_list: filter data ,is list
    :param condition: list,
    :return:
    """
    total = 0
    # print title
    title = ""
    for key in keys.split(","):
        title += "{:<30}".format(key)
    printf("{}".format(title), color="green")

    # print data
    for i in data_list:
        items = ""
        for j in i:
            if len(str(j)) != 0 and isinstance(j, list):
                if isinstance(j[0], str):
                    j = j[0]
                elif isinstance(j[0], dict):
                    j = j[0].get('name', '[unknown]')
                else:
                    j = j
            elif isinstance(j, dict):
                j = j.get('name', '[unknown]')
            else:
                if len(str(j)) == 0:
                    j = '[unknown]'
            j_hex = convert_str(str(j))
            # match to content highlight
            if condition:
                for item in condition:
                    k, v = item.split('=')
                    result_re = re.search(v, j_hex, re.I | re.M)
                    if result_re:
                        # replace to highlight
                        color_content = "\033[31m{}\033[0m".format(result_re.group())
                        j_hex = j_hex.replace(result_re.group(), color_content)
            items += "{:<30}".format(j_hex)
        total += 1
        printf(items)
    print()
    printf("total: {}".format(total))


def print_facets(facets, facet_data, total, figure, table):
    """
    used to print data that users need to count
    :param facets: str, statistical data
    :param facet_data:list, all facets data
    :param total:int, data total
    :param figure:int, data total
    {"app": [{"name": 'xx', 'count': xxx}{"name": 'xx', 'count': xxx}{"name": 'xx', 'count': xxx}]}
    :return:
    """
    if not facet_data:
        return
    # print facet data

    print(' ' + '-' * 40)
    printf(" ZoomEye total data:{}".format(total))
    for facet in facets.split(","):
        names = []
        counts = []
        pie_info = []
        facet_total = {}
        facet_count = 0

        if table.get(facet.strip()) is None:
            support_fields = ','.join(list(table.keys()))
            printf("facet command has unsupport fields [{}], support fields has [{}]".format(facet.strip(),
                                                                                             support_fields),
                   color='red')
            exit(0)
        print(' {:-^40}'.format(facet + " Top 10"))
        if figure is None:
            printf(" {:<35}{:<20}".format(facet, "count"), color="green")
        f = table.get(facet)
        for t in facet_data.get(f):
            facet_count += t.get('count')
        facet_total[facet] = facet_count

        for d in facet_data.get(f):
            name = d.get("name")
            if len(str(name)) == 0:
                name = "[unknown]"

            count = d.get("count")
            # get histogram data
            names.append(name)
            counts.append(count)
            # get pie chart data
            # three decimal places
            pie = (name, round(count / facet_total.get(facet), 3))
            pie_info.append(pie)
            if figure is None:
                printf(" {:<35}{:<20}".format(name, count))
        # pie chart
        if figure and figure == 'pie':
            plotlib.show_pie_chart(pie_info)
        # histogram
        if figure and figure == 'hist':
            plotlib.generate_histogram(counts, names, force_ascii=True)


def print_stat(keys, stat_data, num, figure):
    """
    print current data aggregation
    :param keys: str, field
    :param figure: str, pie or hist
    :param num: int, local total data
    :param stat_data: {'app': {'Gunicorn': 2, 'nginx': 14, 'Apache httpd': 9, '[unknown]': 3, 'Tornado httpd': 2}, 'port': {443: 29, 8443: 1}}
    :return:
    """
    if not stat_data:
        return
    print(' ' + '-' * 40)
    printf(" current total data:{}".format(num), color='green')
    for key in keys.split(','):
        print(' {:-^40}'.format(key + " data"))
        # print title
        if figure is None:
            printf(" {:<35}{:<20}".format(key, "count"), color="green")

        # sort by the amount of each data
        item = stat_data.get(key)
        sorted_item = sorted(item.items(), key=lambda x: x[1], reverse=True)
        # print result
        if figure is None:
            for name, count in sorted_item:
                printf(" {:<35}{:<20}".format(name, count))

        names = []
        counts = []
        pie_info = []
        for name, count in sorted_item:
            names.append(name)
            counts.append(count)
            # get pie chart data
            # three decimal places
            pie = (name, round(count / num, 3))
            pie_info.append(pie)
        # pie chart
        if figure and figure == 'pie':
            plotlib.show_pie_chart(pie_info)
        # histogram
        if figure and figure == 'hist':
            plotlib.generate_histogram(counts, names, force_ascii=True)


def print_host_data(host_data):
    """
    :param host_data, list,
    """
    # host data is None
    if len(host_data) == 0:
        return
    # parser hostname,country,city... information
    first_item = host_data[0]
    all_data, port_count = data.filter_ip_information(data.fields_tables_history_host.keys(),
                                                      data.fields_tables_history_host,
                                                      host_data, omit=True)
    printf(first_item.get('ip'))
    dict_first_item = data.ZoomEyeDict(first_item)
    # print title
    for dict_item in data.tables_history_info.keys():
        result = dict_first_item.find(data.tables_history_info.get(dict_item))
        if result == "" or result is None or result == "Unknown":
            result = "[unknown]"
        printf("{:<30}{}".format(dict_item.capitalize() + ":", result))
    printf("{:30}{}{}".format("Number of open ports:", len(port_count), port_count))
    printf("{:30}{}".format("Number of historical probes:", len(host_data)))
    printf('')
    printf("{:<27}{:<27}{:<27}{:<27}".format("timestamp", "port/service", "app", "banner"), color='green')
    # print detail and process port/service
    for data_item in all_data:
        content = ''
        for item_item in data_item:
            if data_item.index(item_item) == 1:
                res = item_item
                continue
            if data_item.index(item_item) == 2:
                item_item = "{}/{}".format(res, item_item)
            content += "{:<27}".format(item_item)
        printf(content)


def print_filter_history(fileds, hist_data, condition=None):
    """
    print user filter history data,
    :param fileds list,user input field
    :param hist_data dict, from ZoomEye API get data
    :param condition list, filter condition
    """
    filter_title = ''
    first_item = hist_data[0]
    # filter data
    all_data, port_count = data.filter_ip_information(fileds, data.fields_tables_history_host,
                                                      hist_data, omit=False)
    printf(first_item.get('ip'))
    dict_first_item = data.ZoomEyeDict(first_item)
    # parser filter data title
    for dict_item in data.tables_history_info.keys():
        result = dict_first_item.find(data.tables_history_info.get(dict_item))
        if result == "" or result is None or result == "Unknown":
            result = "[unknown]"
        printf("{:<30}{}".format(dict_item.capitalize() + ":", result))
    printf("{:30}{}".format("Number of open ports:", len(port_count)))
    printf("{:30}{}".format("Number of historical probes:", len(hist_data)))
    printf('')
    # print title
    for item in fileds:
        filter_title += "{:<27}".format(item)
    printf(filter_title, color='green')
    # print data
    for data_item in all_data:
        content = ""
        for item_item in data_item:
            # match to content highlight
            if condition:
                for condition_item in condition:
                    k, v = condition_item.split('=')
                    re_result = re.search(str(v), str(item_item), re.I | re.M)
                    # replace to highlight
                    if re_result:
                        content_item = "\033[31m{}\033[0m".format(re_result.group())
                        item_item = str(item_item).replace(re_result.group(), content_item)
            content += "{:<27}".format(item_item)
        printf(content)


def print_information(raw_data):
    if len(raw_data) == 0:
        return
    # parser hostname,country,city... information
    first_item = raw_data[0]
    all_data, port_count = data.filter_ip_information(data.fields_ip.keys(), data.fields_ip, raw_data, omit=True)
    printf(first_item.get('ip'))
    dict_first_item = data.ZoomEyeDict(first_item)
    # print title
    for dict_item in data.tables_history_info.keys():
        result = dict_first_item.find(data.tables_history_info.get(dict_item))
        if result == "" or result is None or result == "Unknown":
            result = "[unknown]"
        printf("{:<30}{}".format(dict_item.capitalize() + ":", result))
    printf("{:30}{}{}".format("Number of open ports:", len(port_count), port_count))
    printf("")
    # print port/service
    printf("{:<10}{:<15}{:<23}{:<30}".format("port", "service", "app", "banner"), color='green')
    for data_item in sorted(all_data, key=lambda k: int(k[0])):
        printf("{:<10}{:<15}{:<23}{:<30}".format(data_item[0], data_item[1], omit_str(data_item[2], 7), data_item[3]))


def print_info_filter(filters, raw_data, condition=None):
    filter_title = ''
    first_item = raw_data[0]
    # filter data
    all_data, port_count = data.filter_ip_information(filters, data.fields_ip,
                                                      raw_data, omit=False)
    printf(first_item.get('ip'))
    dict_first_item = data.ZoomEyeDict(first_item)
    # parser filter data title
    for dict_item in data.tables_history_info.keys():
        result = dict_first_item.find(data.tables_history_info.get(dict_item))
        if result == "" or result is None or result == "Unknown":
            result = "[unknown]"
        printf("{:<30}{}".format(dict_item.capitalize() + ":", result))
    printf("{:30}{}{}".format("Number of open ports:", len(port_count), port_count))
    printf('')
    # print title
    for item in filters:
        filter_title += "{:<27}".format(item)
    printf(filter_title, color='green')
    # print data
    for data_item in all_data:
        content = ""
        for item_item in data_item:
            # match to content highlight
            if condition:
                for condition_item in condition:
                    k, v = condition_item.split('=')
                    re_result = re.search(str(v), str(item_item), re.I | re.M)
                    # replace to highlight
                    if re_result:
                        content_item = "\033[31m{}\033[0m".format(re_result.group())
                        item_item = str(item_item).replace(re_result.group(), content_item)
            content += "{:<27}".format(item_item)
        printf(content)


def show_domain_info(info_list, total, page):
    """
    show query domain info
    Args:
        info_list:

    Returns:
        None
    """

    if len(info_list) == 0:
        return
    printf("{:<55}{:<15}{:<25}".format("name", "timestamp", "ip"), color="green")
    for d in info_list:
        name, timestamp, ip = d.values()
        printf("{:<55}{:<15}{:<25}".format(name, timestamp, ip))
    print()
    printf("total: {}/{}".format(str(30 * int(page)), total))







#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
* Filename: data.py
* Description: used to process various data.
* Time: 2020.11.30
* Author: liuf5
*/
"""
import re
import json
import os
import time
import datetime
import hashlib
from zoomeye import config, file
from zoomeye.sdk import ZoomEye, fields_tables_host, fields_tables_web
from zoomeye.sdk import ZoomEyeDict
from zoomeye import show

zoomeye_dir = os.path.expanduser(config.ZOOMEYE_CONFIG_PATH)

stat_host_table = {
    'app': 'portinfo.app',
    'device': 'portinfo.device',
    'service': 'portinfo.service',
    'os': 'portinfo.os',
    'port': 'portinfo.port',
    'country': 'geoinfo.country.names.en',
    'city': 'geoinfo.city.names.en',
    'ssl': 'ssl'
}

stat_web_table = {
    "webapp": "webapp",
    "component": "component",
    "framework": "framework",
    "server": "server.name",
    "waf": "waf",
    "os": "system",
    "country": "geoinfo.country.names.en",
    "city": "geoinfo.city.names.en",
}

facets_table_host = {
    'app': 'product',
    'device': 'device',
    'service': 'service',
    'os': 'os',
    'port': 'port',
    'country': 'country',
    'city': 'city',
    'ssl': 'ssl'
}

facets_table_web = {
    "webapp": "webapp",
    "component": "component",
    "framework": "framework",
    "server": "server",
    "waf": "waf",
    "os": "os",
    "country": "country",
}

fields_tables_history_host = {
    "timestamp": "timestamp",
    "port": "portinfo.port",
    "service": "portinfo.service",
    "app": "portinfo.product",
    "banner": "raw_data",
    "ssl": "ssl"
}

fields_ip = {
    "port": "portinfo.port",
    "service": "portinfo.service",
    "app": "portinfo.app",
    "banner": "portinfo.banner",
    "ssl": "ssl"
}

tables_history_info = {
    "Hostnames": 'portinfo.hostname',
    'Isp': 'geoinfo.isp',
    "Country": 'geoinfo.country.names.en',
    "City": 'geoinfo.city.names.en',
    "Organization": 'geoinfo.organization',
    "LastUpdated": 'timestamp'
}

default_table_web = {
    "site": "site",
    "title": "title",
    "country": "geoinfo.country.names.en",
    "banner": "raw_data",
}


def md5_convert(string):
    """
    calculate the md5 of a string
    :param string: input string
    :return: md5 string
    """
    m = hashlib.md5()
    m.update(string.encode('utf-8'))
    return m.hexdigest()


def get_host_item(data):
    if data:
        ip = data.get('ip')
        country = data.get("geoinfo").get("country").get("names").get("en")
        port = data.get("portinfo").get("port")
        app = data.get("portinfo").get("app")
        banner = data.get("portinfo").get("banner")
        service = data.get("portinfo").get("service")
        return ip, country, port, app, banner, service
    else:
        show.printf("data cannot be empty", color='red')
        exit(0)


def regexp(keys, field_table, data_list):
    """
    match the corresponding data through regular
    """
    result = []
    for key in keys:
        result = []
        for da in data_list:
            zmdict = ZoomEyeDict(da)
            input_key, input_value = key.split("=")
            if field_table.get(input_key.strip()) is None:
                # check filed effectiveness
                support_fields = ','.join(list(field_table.keys()))
                show.printf("filter command has unsupport fields [{}], support fields has [{}]"
                            .format(input_key, support_fields), color='red')
                exit(0)
            # the value obtained here is of type int, and the user's input is of type str,
            # so it needs to be converted.
            if input_key == "port":
                input_value = str(input_value)
            find_value = zmdict.find(field_table.get(input_key.strip()))
            # get the value through regular matching
            try:
                regexp_result = re.search(str(input_value), str(find_value), re.I)
            except re.error:
                show.printf('the regular expression you entered is incorrect, please check!', color='red')
                exit(0)
            except Exception as e:
                show.printf(e, color='red')
                exit(0)
            # the matched value is neither None nor empty
            if regexp_result and regexp_result.group(0) != '':
                result.append(da)
        # AND operation
        data_list = result
    return result


def filter_search_data(keys, field_table, data):
    """
    get the data of the corresponding field
    :param keys: list, user input field
    :param field_table: dict, fileds
    :param data: list, zoomeye api data
    :return: list, ex: [[1,2,3...],[1,2,3...],[1,2,3...]...]
    """
    result = []
    for d in data:
        item = []
        zmdict = ZoomEyeDict(d)
        for key in keys:
            if field_table.get(key.strip()) is None:
                support_fields = ','.join(list(field_table.keys()))
                show.printf("filter command has unsupport fields [{}], support fields has [{}]"
                            .format(key, support_fields), color='red')
                exit(0)
            res = zmdict.find(field_table.get(key.strip()))
            if key == "timestamp":
                utc_time = datetime.datetime.strptime(res, "%Y-%m-%dT%H:%M:%S")
                res = str(utc_time + datetime.timedelta(hours=8))
            item.append(res)
        result.append(item)
    return result


def filter_ip_information(fileds, tables, host_data, omit=True):
    """
    filter historical data based on user input
    :param fileds: list, user input
    :param host_data: list, exclude web data
    :param omit: bool, omit string flag
    return: all_data,list matched data
    return: port_count, set, count open ports non-repeating
    """
    all_data = []
    port_count = set()

    for host_item in host_data:
        every_item = []
        host_dict = ZoomEyeDict(host_item)
        for filed_item in fileds:
            host_result = host_dict.find(tables.get(filed_item.strip()))
            # count ports
            if filed_item == 'port':
                port_count.add(host_result)
            # format timestamp
            if filed_item == 'timestamp':
                utc_time = datetime.datetime.strptime(host_result[:19], "%Y-%m-%dT%H:%M:%S")
                host_result = str(utc_time)
            # omit raw data, is too long
            if filed_item == 'banner':
                if omit:
                    host_result = show.omit_str(show.convert_str(host_result))
                else:
                    host_result = show.convert_str(host_result)
            if filed_item.lower() == 'ssl':
                host_result = host_dict.find(tables.get(filed_item.lower().strip()))
            # replace None --> [unknown]
            if host_result is None:
                host_result = "[unknown]"
            every_item.append(host_result)
        all_data.append(every_item)
    return all_data, port_count


def process_filter(fields, data, tables):
    has_equal = []
    not_equal = []

    if not data:
        return
    # resolve specific filter items
    for field_item in fields:
        field_split = field_item.split('=')
        if len(field_split) == 2:
            has_equal.append(field_item)
            # not_equal.append(field_split[0])
        if len(field_split) == 1:
            if field_item == "*":
                not_equal = list(tables.keys())
                continue
            else:
                not_equal.append(field_item)
    # match filters that contain specific data
    if len(has_equal) != 0:
        result_data = regexp(has_equal, tables, data)
    else:
        result_data = data

    return result_data, has_equal, not_equal


class Cache:
    """
    used to cache the data obtained from the api to the local,
    or directly clip the file from the local.
    """

    def __init__(self, dork, resource, page=0):
        self.dork = dork
        self.page = page
        self.resource = resource
        self.cache_folder = os.path.expanduser(config.ZOOMEYE_CACHE_PATH)
        self.filename = self.cache_folder + config.FILE_NAME.format(md5_convert(dork + self.resource), page)

    def check(self):
        """
        used to detect whether there is a local cache file,
        and check whether the local cache file expires.
        :return: bool
        """
        # exist?
        if not os.path.exists(self.filename):
            return False
        # check expired
        file_create_time = os.path.getatime(self.filename)
        # expired time is five day, time stamp = 60 * 60 * 24 * 5
        if int(time.time()) - int(file_create_time) > config.EXPIRED_TIME:
            # delete expired cache file
            os.remove(self.filename)
            return False
        # the cache file exists and has not expired
        else:
            return True

    def save(self, data):
        # folder is exists?
        if not os.path.isdir(self.cache_folder):
            try:
                # create folder when folder does not exists
                os.makedirs(self.cache_folder)
            except OSError:
                raise Exception("init dirs failed {}".format(self.cache_folder))
        # write to local
        with open(self.filename, 'w') as f:
            f.write(data)

    def load(self):
        """
        files in the cache from the local folder
        :return:
        """
        result_data = []
        facet_data = None
        with open(self.filename, 'r') as f:
            read_data = f.read()
        data_json = json.loads(read_data)
        """
        process the data loaded from the cache file, 
        the processed data structure is list. like:
        [{...}, {...},{...}...]
        """
        for i in data_json.get("matches"):
            result_data.append(i)
        # get facet
        if data_json.get("facets"):
            facet_data = data_json.get("facets")
        # get data total
        total = data_json.get("total")
        # return
        return result_data, facet_data, total

    def update(self):
        pass


class CliZoomEye:

    def __init__(self, dork, num, resource, facet=None, force=False):
        self.dork = dork
        self.num = num
        self.resource = resource
        self.facet = facet
        self.force = force

        self.dork_data = list()
        self.facet_data = None
        self.total = 0

        self.api_key, self.access_token = file.get_auth_key()
        self.zoomeye = ZoomEye(api_key=self.api_key, access_token=self.access_token)

    def handle_page(self):
        try:
            num = int(self.num)
            if num % 20 == 0:
                return int(num / 20)
            return int(num / 20) + 1
        except ValueError:
            if self.num == 'all':
                for i in range(3):
                    user_confirm = input("The data may exceed your quota. "
                                         "If the quota is exceeded, all quota data will be returned. "
                                         "Are you sure you want to get all the data?(y/N)\n")
                    if user_confirm == 'y':
                        self.zoomeye.dork_search(
                            dork=self.dork,
                            page=1,
                            resource=self.resource,
                            facets=self.facet
                        )
                        self.num = self.zoomeye.total
                        cache = Cache(self.dork, page=1, resource=self.resource)
                        cache.save(json.dumps(self.zoomeye.raw_data))
                        if self.num % 20 == 0:
                            return int(self.num / 20)
                        return int(self.num / 20) + 1
                    elif user_confirm == "N":
                        user_num = input("Please enter the required amount of data:")
                        return int(user_num)
                    else:
                        continue
                show.printf("more than the number of errors!", color='red')
        except Exception as e:
            show.printf(e, color='red')
            exit(0)

    def cache_dork(self, page, data):
        cache_file = Cache(self.dork, page=page, resource=self.resource)
        cache_file.save(json.dumps(data))

    def request_data(self):
        if os.path.exists(self.dork):
            self.load()
        else:
            page_count = self.handle_page()
            for page in range(page_count):
                cache_file = Cache(self.dork, self.resource, page)
                if cache_file.check() and self.force is False:
                    dork_data_list, self.facet_data, self.total = cache_file.load()
                    self.dork_data.extend(dork_data_list)
                else:
                    if self.resource == 'host':
                        self.facet = ['app', 'device', 'service', 'os', 'port', 'country', 'city']
                    if self.resource == 'web':
                        self.facet = ['webapp', 'component', 'framework', 'frontend',
                                      'server', 'waf', 'os', 'country', 'city']
                    try:
                        dork_data_list = self.zoomeye.dork_search(
                            dork=self.dork,
                            page=page + 1,
                            resource=self.resource,
                            facets=self.facet
                        )
                    except ValueError:
                        print("the access token expires, please re-run [zoomeye init] command."
                              "it is recommended to use API KEY for initialization!")
                        exit(0)
                    self.facet_data = self.zoomeye.facet_data
                    self.total = self.zoomeye.total
                    self.dork_data.extend(dork_data_list)
                    self.cache_dork(page, self.zoomeye.raw_data)

    def default_show(self):
        self.request_data()
        if self.resource == 'host':
            show.show_host_default_data(self.dork_data, self.total)
        if self.resource == 'web':
            show.show_web_default_data(self.dork_data, self.total)

    def filter_data(self, keys, save=False):
        """
        according to web/search or host/search select filter field
        """
        if save is not True:
            self.request_data()
        if self.resource == 'host':
            tables = fields_tables_host
        if self.resource == 'web':
            tables = fields_tables_web
        has_equal = []
        not_equal = []
        # set the ip field to be displayed by default and in the first place
        key_list = keys.split(',')
        try:
            # set ip field in the first place
            key_index = key_list.index("ip")
            key_list.pop(key_index)
            key_list.insert(0, 'ip')
        # add IP to the first item when there is no IP
        except ValueError:
            key_list.insert(0, 'ip')
        # process user input fields, separating single fields from fields with equal signs.
        for key in key_list:
            res = key.split('=')
            # field with equal sign
            if len(res) == 2:
                has_equal.append(key)
                # not_equal.append(res[0])
            # No field with equal sign
            if len(res) == 1:
                # handle the case where the wildcard character * is included in the filed
                # that is, query all fields
                if key == "*":
                    not_equal = list(tables.keys())
                    continue
                else:
                    not_equal.append(key)
        # the filter condition is port, banner, app=**
        # ex:port,banner,app=MySQL
        if len(has_equal) != 0:
            equal_data = regexp(has_equal, tables, self.dork_data)
        # the filter condition is app, port
        # ex: ip,port,app
        else:
            equal_data = self.dork_data
        # get result
        result = filter_search_data(not_equal, tables, equal_data)
        equal = ','.join(not_equal)
        if save:
            return equal, result
        show.print_filter(equal, result[:self.num], has_equal)

    def facets_data(self, facet, figure):
        """
        according to web/search or host/search select facet field
        """
        self.request_data()
        if self.resource == 'host':
            tables = facets_table_host
        if self.resource == 'web':
            tables = facets_table_web
        show.print_facets(facet, self.facet_data, self.total, figure, tables)

    def save(self, fields):
        """
        save api response raw data
        """
        self.request_data()
        name = re.findall(r"[a-zA-Z0-9_\u4e00-\u9fa5]+", self.dork)
        re_name = '_'.join(name)

        if fields == 'all':
            filename = "{}_{}_{}_{}.json".format(re_name, self.num, self.resource, int(time.time()))
            data = {
                'total': self.total,
                'matches': self.dork_data,
                'facets': self.facet_data
            }
            with open(filename, 'w') as f:
                f.write(json.dumps(data))
            show.printf("save file to {}/{} successful!".format(os.getcwd(), filename), color='green')
        # -save xx=xxxx, save the filtered data. data format ex:
        # {app:"xxx", "app":"httpd",.....}
        else:
            key, value = self.filter_data(fields, save=True)
            filename = "{}_{}_{}.json".format(re_name, len(value), int(time.time()))
            # parser data
            for v in value:
                dict_save = {}
                for index in range(len(key.split(','))):
                    dict_key = key.split(',')[index]
                    if isinstance(v[index], dict):
                        v[index] = v[index].get('name', 'unknown')
                    dict_value = v[index]
                    dict_save[dict_key] = dict_value
                # write to local file
                file.add_to_file(filename, str(dict_save))
            show.printf("save file to {}/{} successful!".format(os.getcwd(), filename), color='green')

    def load(self):
        """
        load local json file,must be save file
        """
        with open(self.dork, 'r') as f:
            data = f.read()
        try:
            json_data = json.loads(data)
        except json.decoder.JSONDecodeError:
            show.printf('json format error', color='red')
            exit(0)
        self.total = json_data.get("total", 0)
        self.dork_data = json_data.get("matches", "")
        self.facet_data = json_data.get("facets", "")
        self.num = len(self.dork_data)

    def statistics(self, keys, figure):
        """
        local data statistics
        """
        self.request_data()
        if self.resource == 'web':
            tables = stat_web_table
        if self.resource == 'host':
            tables = stat_host_table
        data = {}
        key_list = keys.split(',')
        # cycle key
        for key in key_list:
            count = {}
            for item in self.dork_data[:self.num]:
                zmdict = ZoomEyeDict(item)
                if tables.get(key.strip()) is None:
                    # check filed effectiveness
                    support_fields = ','.join(list(tables.keys()))
                    show.printf("filter command has unsupport fields [{}], support fields has [{}]"
                                .format(key, support_fields), color='red')
                    exit(0)
                fields = zmdict.find(tables.get(key.strip()))
                # the value of the result field returned by the API may be empty
                if fields == '' or isinstance(fields, list):
                    fields = '[unknown]'
                r = count.get(fields)
                if not r:
                    count[fields] = 1
                else:
                    count[fields] = count[fields] + 1
            data[key] = count
        # print result for current data aggregation
        show.print_stat(keys, data, self.num, figure)

    def count(self):
        """
        show dock count number
        """
        self.request_data()
        show.printf(self.total)


class HistoryDevice:
    """
    obtain the user's identity information and determine whether to use the IP history search function
    """

    def __init__(self, ip, force, num):
        self.ip = ip
        self.force = force
        self.num = num
        self.cache_path = os.path.expanduser(config.ZOOMEYE_CACHE_PATH) + "/" + self.ip

    def cache_data(self, history_data):
        """
        save ip history data to local
        :param history_data: dict, ip history data
        """
        try:
            # write data
            # if file path not exists
            with open(self.cache_path, 'w') as f:
                f.write(json.dumps(history_data))
        except FileNotFoundError:
            # create fold
            # retry write to local file
            os.makedirs(os.path.expanduser(config.ZOOMEYE_CACHE_PATH))
            with open(self.cache_path, 'w') as f:
                f.write(json.dumps(history_data))
        except Exception as e:
            show.printf("unknown error: {}".format(e))
            exit(0)

    def get_data_from_cache(self):
        """
        get data from local file
        """
        # cache file exists
        if os.path.exists(self.cache_path):
            # file exists check expired time
            # expired time five day
            create_time = os.path.getatime(self.cache_path)
            if int(time.time()) - int(create_time) > config.EXPIRED_TIME:
                # over expired time remove file
                os.remove(self.cache_path)
                return None
            # read local file
            with open(self.cache_path, 'r') as f:
                history_data = f.read()
            return history_data
        # cache file not exists
        else:
            return None

    def drop_web_data(self):
        host_data = []
        not_split_data = self.get_data()
        for item in not_split_data.get('data'):
            if "component" in item.keys():
                continue
            host_data.append(item)
        return host_data

    def get_data(self):
        """
        get user level and IP historical data
        """
        normal_user = ['user', 'developer']
        api_key, access_token = file.get_auth_key()
        zm = ZoomEye(api_key=api_key, access_token=access_token)
        role = zm.resources_info()
        # permission restrictions
        if role["plan"] in normal_user:
            show.printf("this function is only open to advanced users and VIP users.", color='red')
            exit(0)
        # the user chooses to force data from the API
        if self.force:
            history_data = zm.history_ip(self.ip)
        else:
            # from local cache get data
            history_data_str = self.get_data_from_cache()
            # local cache not exists from API get data
            if history_data_str is None:
                history_data = zm.history_ip(self.ip)
            else:
                history_data = json.loads(history_data_str)
        # cache data
        self.cache_data(history_data)
        return history_data

    def show_fields(self):
        """
        print some field in terminal
        """
        host_data = self.drop_web_data()
        if self.num is not None and self.num >= len(host_data):
            self.num = len(host_data)
        show.print_host_data(host_data[:self.num])

    def filter_fields(self, fields):
        """
        filter historical IP data
        :param fields: list, user input filter fields
        """
        data = self.drop_web_data()
        result_data, has_equal, not_equal = process_filter(fields, data, fields_tables_history_host)
        # no regexp data
        if len(result_data) == 0:
            return
        # check user input filed is or not support
        for item in not_equal:
            if fields_tables_history_host.get(item.strip()) is None:
                support_fields = ','.join(list(fields_tables_history_host.keys()))
                show.printf(
                    "filter command has unsupport fields [{}], support fields has [{}]".format(item, support_fields),
                    color='red')
                exit(0)
        show.print_filter_history(not_equal, result_data[:self.num], has_equal)


class IPInformation:
    """
    query IP information
    """

    def __init__(self, dork):
        self.dork = "ip:{}".format(dork)

    def request_data(self):
        """
        get api data
        """
        api_key, access_token = file.get_auth_key()
        zm = ZoomEye(api_key=api_key, access_token=access_token)
        data = zm.dork_search(self.dork)
        return data

    def show_information(self):
        """
        show default fields: port,service,app,banner
        """
        info_data = self.request_data()
        show.print_information(info_data)

    def filter_information(self, filters):
        """
        filter IP information, filter like key or key,key=value
        """
        info_data = self.request_data()
        result_data, has_equal, not_equal = process_filter(filters, info_data, fields_ip)
        if len(result_data) == 0:
            return
        for item in not_equal:
            if fields_ip.get(item.strip()) is None:
                support_fields = ','.join(list(fields_ip.keys()))
                show.printf(
                    "filter command has unsupport fields [{}], support fields has [{}]".format(item, support_fields),
                    color='red')
                exit(0)
        show.print_info_filter(not_equal, result_data, has_equal)


class DomainSearch:
    """
    query relation domain or sub domain
    """

    def __init__(self, q, source, page):
        self.q = q
        self.source = source
        self.page = page
        api_key, access_token = file.get_auth_key()
        self.zm = ZoomEye(api_key=api_key, access_token=access_token)

    def show_information(self):
        """show domain search data"""

        info_data, total = self.zm.domain_search(self.q, self.source, self.page)
        show.show_domain_info(info_data, total, self.page)
        # return None




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
* Filename: __init__.py.py
* Description:
* Time: 2020.11.25
* Author: liuf5
*/
"""

__name__ = 'zoomeye'
__package__ = 'zoomeye'
__version__ = ' 2.0.4.6.11'
__site__ = "https://www.zoomeye.org/doc"
